use core::{fmt::Display, str};

// TODO: we don't need itertools once https://github.com/rust-lang/rust/issues/79524 lands
use itertools::Itertools;

/// The "hello" packet to start communication with the ATEM
pub const COMMAND_CONNECT_HELLO: [u8; 20] = [
    0x10, 0x14, 0x53, 0xab, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3a, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
];

pub struct AtemPacket<T: AsRef<[u8]>> {
    buf: T,
}

#[derive(Debug)]
pub enum AtemPacketErr {
    /// The packet was too short
    TooShort { got: usize },
    /// The packet's stated and actual lengths were different
    LengthDiffers { expected: u16, got: usize },
}

#[derive(PartialEq)]
pub enum PacketFlag {
    AckRequest,
    NewSessionId,
    IsRetransmit,
    RetransmitRequest,
    AckReply,
}

impl From<PacketFlag> for u8 {
    fn from(flag: PacketFlag) -> Self {
        match flag {
            PacketFlag::AckRequest => 0x01,
            PacketFlag::NewSessionId => 0x02,
            PacketFlag::IsRetransmit => 0x04,
            PacketFlag::RetransmitRequest => 0x08,
            PacketFlag::AckReply => 0x10,
        }
    }
}

impl<T: AsRef<[u8]>> AtemPacket<T> {
    pub fn new_checked(buf: T) -> Result<Self, AtemPacketErr> {
        let len = buf.as_ref().len();
        if len < 12 {
            return Err(AtemPacketErr::TooShort {
                got: buf.as_ref().len(),
            });
        }
        let p = Self { buf };
        if p.length() as usize != len {
            return Err(AtemPacketErr::LengthDiffers {
                expected: p.length(),
                got: len,
            });
        }
        Ok(p)
    }
    pub fn length(&self) -> u16 {
        u16::from_be_bytes(self.buf.as_ref()[0..=1].try_into().unwrap()) & 0x07ff
    }

    pub fn flags(&self) -> u8 {
        self.buf.as_ref()[0] >> 3
    }

    pub fn session_id(&self) -> u16 {
        u16::from_be_bytes(self.buf.as_ref()[2..=3].try_into().unwrap())
    }

    pub fn ack_number(&self) -> u16 {
        u16::from_be_bytes(self.buf.as_ref()[4..=5].try_into().unwrap())
    }

    pub fn remote_sequence_number(&self) -> u16 {
        u16::from_be_bytes(self.buf.as_ref()[9..=10].try_into().unwrap())
    }

    pub fn remote_packet_id(&self) -> u16 {
        u16::from_be_bytes(self.buf.as_ref()[10..=11].try_into().unwrap())
    }

    /// Return true if this packet has the given [`PacketFlag`]
    pub fn has_flag(&self, flag: PacketFlag) -> bool {
        self.flags() & u8::from(flag) > 0
    }

    /// Get an iterator over the `Field`s in this packet.
    ///
    /// Returns None if this is a packet without fields.
    pub fn fields(&self) -> Option<Fields> {
        // TODO: do we only ever get newsessionid during the handshake (i.e. not in a packet with fields)?
        if self.has_flag(PacketFlag::NewSessionId) {
            None
        } else {
            Some(Fields {
                data: self.body(),
                offset: 0,
            })
        }
    }

    pub fn body(&self) -> &[u8] {
        &self.buf.as_ref()[12..]
    }
}

#[cfg(feature = "std")]
impl<T: AsRef<[u8]>> Display for AtemPacket<T> {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        write!(
            f,
            "sid: {}, len: {}, fields: ({})",
            self.session_id(),
            self.length(),
            self.fields()
                .and_then(|f| Some(
                    f.map(|f| str::from_utf8(f.r#type).unwrap())
                        .intersperse(", ")
                        .collect::<String>()
                ))
                .unwrap_or("none".into())
        )
    }
}

/// An ATEM protocol field - a 4-ascii character type plus variable length data
pub struct Field<'a> {
    r#type: &'a [u8; 4],
    data: &'a [u8],
}

/// Created by [`AtemPacket::fields`]
pub struct Fields<'a> {
    data: &'a [u8],
    offset: usize,
}

impl<'a> Iterator for Fields<'a> {
    type Item = Field<'a>;
    fn next(&mut self) -> Option<Self::Item> {
        let remain = self.data.len() - self.offset;
        if remain == 0 {
            return None;
        } else if remain < 8 {
            // TODO: is 8 indeed the minimum size for something here? (i.e. no field data)
            panic!("Oh no");
        }

        let length =
            u16::from_be_bytes(self.data[self.offset..=self.offset + 1].try_into().unwrap());
        // TODO: sanity check length
        let r#type: &[u8; 4] = self.data[self.offset + 4..=self.offset + 7]
            .try_into()
            .unwrap();
        let data = &self.data[self.offset + 8..self.offset + (length as usize)];
        self.offset += (length as usize);
        Some(Field { r#type, data })
    }
}
